#!/usr/local/bin/perl -w

use strict;
use MySubst;

sub _html_escape
{
	my $text = shift;
    return undef unless defined $text;
	$text =~ s/</\&#60;/g ;
	$text =~ s/>/\&#62;/g ;
	$text =~ s/"/\&#34;/g ;
	return $text;
}

# Makes an HTML link,
# given URI, text, and optionally a target window name.
# The URI is assumed to be valid.
#
sub mk_link {
	my($href,$text,$target_window) = @_;
	my $rv = '<a href="' . $href . '">'
				  . _html_escape($text)
				  . '</a>';
				  ;
	if ($target_window) {
		$rv =~ s/>/ target="$target_window">/;
	}
	return $rv;
}

# returns the contents of a file,
# given the file name or readable file handle
#
sub getfile {
	my $name = shift;
	my $fh;
	if (ref $name) {
		$fh = $name;
	} else {
		if (!open($fh, "<$name" )) {
			#printf STDERR "%s\n", "can't open '$name': $!";
			my $msg = sprintf "%s\n", "can't open '$name': $!";
			die $msg;
		}
	}
	my $sav = $/;
	$/ = undef;
	my $var = <$fh>;
	$/ = $sav;
	close $fh;
	$var;
}

# globals; can be modified via the command line

my $debug = 1;
my $infile = 'photos.txt'; # input file describing the images
my $pagelist = "pages.txt"; # the output file for the list of pages
my $pgdir = '.'; # location of the HTML fragments for each page
my $serve_dir = 'photos'; # where the server and script find everything
my $mdir = 'm/'; # location of the intermediate-sized image files
my $convert_geom = '350x350>'; # geometry of the intermediate image files
my $bigfiles = 1; # if 0, no full-size image files on server
my $no_create = 0; # if 1, do not create the intermediate image files
my $menu_width = '20%'; # menu width; can be overriden by command line
my $menu_width_cl = ''; # value given on command line
my $complete_html=0;
my $page_template_file = '';
my $page_template = '';
my $page_suffix; # undefined by default
my $img_template_file = '';
my $img_template = '';
my $href_template_file = '';
my $href_template = '';

my $year = (localtime())[5] + 1900;
my $copyright = "All images Copyright &copy; " . $year;

# can be changed via command line; see below
my $crlink =
	mk_link("http://thunderingpaws.org", "Thundering Paws", "_blank");

## other globals
my %alias = ();
my @rings = ();
my @ex_auto = ();
my %ex_autolinks = ();
my $br = '<br>' . "\n";
my $subst = new MySubst;


# command line

for my $arg (@ARGV) {
	if ($arg =~ /^infile=(\S+)/) {
		$infile = $1;
	}
	elsif ($arg =~ /^pgdir=(\S+)/) {
		$pgdir = $1;
	}
	elsif ($arg =~ /^mdir=(\S+)/) {
		$mdir = $1;
		$mdir .= '/' if $mdir !~ qr|/$|;
	}
	elsif ($arg =~ /^pagelist=(\S+)/) {
		$pagelist = $1;
	}
	elsif ($arg =~ /^serve_dir=(\S+)/) {
		$serve_dir = $1;
	}
	elsif ($arg =~ /^crlink=(\S+)/) {
		$crlink = $1;
	}
	elsif ($arg =~ /^geom=(\S+)/) {
		$convert_geom = $1;
	}
	elsif ($arg =~ /^bigfiles=(\S+)/) {
		$bigfiles = $1;
	}
	elsif ($arg =~ /^no_create=(\S+)/) {
		$no_create = $1;
	}
	elsif ($arg =~ /^menu_width=(\S+)/) {
		$menu_width = $1;
		$menu_width_cl = $1;
	}
	elsif ($arg =~ /^complete_html=(\S+)/) {
		$complete_html = $1; # true = create complete HTML files
	}
	elsif ($arg =~ /^page_template=(\S+)/) {
		$page_template_file = $1; # template for the pages
	}
	elsif ($arg =~ /^img_template=(\S+)/) {
		$img_template_file = $1; # template for the <img ...> tags
	}
	elsif ($arg =~ /^href_template=(\S+)/) {
		$href_template_file = $1; # template for the <img ...> tags
	}
	elsif ($arg =~ /^page_suffix=(\S+)/) {
		$page_suffix = $1; # suffix of the output pages
	}
}

$copyright .= ' by ' . $crlink ;

$convert_geom =~ /([0-9]+)x([0-9]+)/ or die "bad geom '$convert_geom'";
my $convert_geom_w = $1;
my $convert_geom_h = $2;


## Functions for generating parts of the HTML files
##

sub mk_page_entry_text {
	my($name,$file,$title,$id) = @_;
	my $rv =
		$name
		. "|" . $file
		. "|" . ($title || "")
		. ($id ? ("|" . $id) : "")
		;
	return $rv;
}

# default templates
#
sub photo_page_template {

return <<\TEMPLATE_END;
<%no menu%><table width="100%"><tr><td valign="top" align="left" width="%%pic_menu_width%%">
%%pic_menu%%
</td>
<td valign="top" width="%%pic_width%%">
%%img%%
</td><td valign="top">
%%pic_description%%
</td></tr></table>
TEMPLATE_END

}

sub img_template {
return '<img src="%%pic_url%%" width="%%pic_width%%" height="%%pic_height%%"
 alt="%%pic_name%%" title="%%pic_name%%" class="album">'
}

sub img_template2 {
return 
 '<a target="_blank"'
			. ' href="%%pic_bigfile_url%%"'
			. ' style="border: 0;">'
			. img_template()
			. '</a>'
}

# Create an HTML page containing a picture
# Input: A ref to a hash of substitutions
#        The keys are 
#        pic_url
#        pic_width
#        pic_height
#        pic_bigfile_url (may be undefined)
#        pic_menu
#        pic_menu_width
#        pic_name
#        pic_gender
#        pic_dob
#        pic_record_date
#        pic_description
#        
#
sub create_photo_page {
	my($values, $page_template, $img_template)=@_;

	#my $pic_width = $$values{pic_width} || '';
	#my $sb = mk_pic_sandbags($pic_width);
	#$$values{pic_sandbags} = $sb;

	my $imgtag = $img_template
		|| ( $$values{pic_bigfile_url} 
	             ? img_template2()
				 : img_template() );
	$$values{pic_menu_width} = '30%' if ! $$values{pic_menu_width};

	my $page = $page_template ? $page_template : photo_page_template();

	if (!$subst) {
		print STDERR "creating MySubst object\n";
		$subst = new MySubst;
	}
	$page =  $subst->subst($page,{img=>$imgtag});
	my $rv = $subst->subst($page,$values);
	return $rv;
}

##

my $wholefile = getfile($infile);
my @sections = split /^%/m, $wholefile;

$page_template = getfile($page_template_file) if $page_template_file;
$img_template = getfile($img_template_file) if $img_template_file;
if ($href_template_file) {
	$href_template = getfile($href_template_file);
	$href_template =~ s/\s+$//;
}

my %cats; # not currently used?
my @images = ();
my %titles = ();
my @menu_end = ();
my @menu_begin = ();
my $suffix = ".txt";
if (defined($page_suffix)) {
	$suffix = $page_suffix;
} elsif ($complete_html || $page_template) {
	$suffix = ".htm";
}

for my $sec (@sections) {
	my @lines;
	@lines = split /\n/m, $sec;
	next if ! @lines;
	my $line0 = shift @lines;
	@lines = map { /^#/ ? () : ($_) } @lines; # drop comment lines
	next if ! @lines;
	if ($line0 =~ /^cats/) { ## unused?
		for my $l (@lines) {
			next if $l !~ /^[a-zA-Z]/;
			my @fff = split /,/, $l;
			my $nm = $fff[0];
			$cats{$nm} = {name=>$nm,
			             adoptable=>$fff[1],
			             birthdate=>$fff[2],
						 };
		}
	} elsif ($line0 =~ /^images/) {
		my $block = join("\n",@lines)."\n";
		my @rec = split /^[.][^\n]*\n/m, $block;
		my $prop_pattern = qr/^(\w+)=(.*)/;
		for my $r (@rec) {
			$r =~ s/^\s+//s;
			#$r =~ s/\s+$//s;
			next if $r =~ /^\s*$/s;
			my $num = 1 + @images;
			my $outfile = $num . $suffix;
			my @fff = split /[|]/, $r;
			my $img = {};

			## Assign fields given like "menu=0"
			print STDERR "---\n" if $debug;
			@fff = map {
				if ( $_ =~ $prop_pattern ) {
					$$img{$1} = $2; 
					print STDERR "$1=$2\n" if $debug;
					();
				} else { ($_) }
			} @fff;

			## Assign other fields
			my $title = $fff[1];
			printf STDERR "title=%s\n", $title;
			my %props = (filename=>$fff[0],
			               title=>$title,
						   html_description=>$fff[2],
						   number=>$num,
						   pgfile=>$outfile,
						   rings=>[]
						   );
			while (my($key,$value) = each %props) {
            	print STDERR "$key=$value\n" if $debug;
				$$img{$key} = $value;
            }
			print STDERR "---\n" if $debug;

			push @images, $img;

			## remember the first image with this title
			$titles{$fff[1]} = $img
				if !defined( $titles{$fff[1]} );
		}
	} elsif ($line0 =~ /^alias/) { ## aliases for titles (automatic links)
		for my $l (@lines) {
			next if $l !~ /^[a-zA-Z]/;
			my @fff = split /\s*=\s*/, $l;
			$alias{$fff[0]} = $fff[1];
		}

	} elsif ($line0 =~ /^autolink/) { ## more automatic links

		for my $l (@lines) {
			next if $l =~ /^\s*#/;
			my @fff = split /\s*=\s*/, $l, 2;
			my ($tok,$url) = @fff;
			my $window = '';
			$window = '_blank' if $tok =~ s/^\+//;
			push @ex_auto, $tok;
			$ex_autolinks{$tok} = {url=>$url, window=>$window};
		}
		

	} elsif ($line0 =~ /^ring/) { ## pages linked in rings
		for my $l (@lines) {
			next if $l !~ /^[a-zA-Z]/;
			my @fff = split /\s*:\s*/, $l, 2;
			push @rings, {text=>$fff[0],
			              pattern=>$fff[1],
						  images=>[]
						  };
		}

	} elsif ($line0 =~ /^menu/) { ## additions to the menu
		for my $l (@lines) {
			next if $l !~ /^[a-zA-Z]/;
			my @fff = split /\|/, $l;
			my $type = $fff[0];
			if ($type eq 'width') {
				if (! $menu_width_cl) {
					$menu_width = $fff[1];
				}
				next;
			}
			my $rec = { name=>$fff[1],
						url=>$fff[2],
						title=>$fff[3],
						};
			if ($type eq 'add-end') {
				push @menu_end, $rec;
			} elsif ($type eq 'add-begin') {
				push @menu_begin, $rec;
			}
		}
	}
}

printf STDERR "cats:%s\n\n", join(",",%cats) if $debug;
printf STDERR "images:%s\n\n", join(",",@images) if $debug;
printf STDERR "menu_end:%s\n\n", join(",",@menu_end) if $debug;

# &mk_href creates an "href" (link destination),
# given an an image record
#
# If no href_template was given, the return value is
#	%%url%%?/serve_dir,NN
# The %%url%% is a literal substring, meant to be replaced later.
# The serve_dir is the server directory, which can be given on command line.
# The NN is the sequential image number.
# 
#
my $prepend_script_url = 0;
if (!$href_template) {
	$href_template = $serve_dir . ',%%number%%';
	$prepend_script_url = 1;
}
# sub mk_href {
# 	my $img = $_[0];
# 	my $i = $$img{number};
# 	my $href = $href_template;
# 	$href =~ s/{}/$i/;
# 	return $href;
# }
sub mk_href {
	my $img = $_[0];
	my $href = $subst->subst($href_template, $img); 
	if ($prepend_script_url) {
		$href = '%%url%%?/' . $href;
	}
	return $href;
}


## Given a page title, or other string,
## return a link to the page to which the string refers.
##
sub title_to_link {
	my($title) = @_;
	my $rv;

	my $ex = $ex_autolinks{$title};  ## explicitly defined link
	my $window = '';
	my $url;
	if (defined($ex)) {
		($url,$window) = ($$ex{url}, $$ex{window});
	} else {
		## link to one of the pages we are creating
		##
		my $img = $titles{$title} || $titles{$alias{$title}};
		die "no image titled '$title'" if ! $img;
		$url = mk_href($img);
	}
	$rv = mk_link($url, $title, $window);

	return $rv;
}

# Replaces all occurrences of the given pattern in the input
# text with the result of passing the matched text to
# a subroutine.
# 
# E.g. if pattern is "cat|dog" and subr is "sub {'('.$_[0].')'}"
# then the word "cat" in the input is replaced with
# "(cat)", and "dog" is replaced with "(dog)".
# The new text is returned.
# 
# Does not replace parts of an input word, e.g. if the pattern
# is "cat|dog" then "cathode" and "muddog" are left unchanged.
#
sub subst_word {
	my($text, $subr, $ptn) = @_;

	$text =~ s{^($ptn)(?=\W|$)
	          |(?<=\W)($ptn)(?=\W|$)
			  }
			  {&$subr($&)}exg;

	return $text;
}

# Make a hyperlink.
# If the URL is prefixed with '+',
# remove the '+' and make a link to open in a new window
#
sub mk_link_plus {
	my($href,$text) = @_;
	my @target = ();
	if ( substr($href,0,1) eq '+') {
		$href = substr($href,1);
		@target = ('_blank');
	}

	return mk_link($href,$text,@target)
}

sub createfiles {
	my $img = shift;
	my $big = $$img{filename};
	my $tt = $$img{title};

	# top part of the menu
	#
	my $menu = '<div class="menu2">' . "\n";
	my $next = $images[ $$img{number} % scalar(@images) ];
	my $prev = $images[ ($$img{number}-2) % scalar(@images) ];
	my $href;
	$menu .= '<div style="border-bottom: 1px dotted black">';
	$menu .= '<table width="100%"><tr><td align="center">';
	$href = mk_href($prev);
	$menu .= mk_link($href,' &nbsp;<<<<&nbsp; ') . "\n";
	$menu .= qq{</td>\n<td align="center">};
	$href = mk_href($next);
	$menu .= mk_link($href,' &nbsp;>>>>&nbsp; ') . "\n";
	$menu .= '</td></tr></table>';

	for my $aa (@menu_begin) {
		$menu .= mk_link_plus($$aa{url},$$aa{title}) . "\n";
	}
	$menu .= '</div>';

	# main part of the menu.
	# also, create a pattern of all the other image titles.
	#
	my $ptn = '';
	my @autolink = @ex_auto;
	for my $ii (@images) {
		if (defined ($$ii{menu})) {
			print STDERR "title:$$ii{title},menu=$$ii{menu}\n"
				if $debug;
			next if $$ii{menu} =~ /^(?i:0|no|exclude)$/ ;
		}
		my $text = $$ii{title};
		if ($ii == $img) {
			$text = '[ ' . $text . ' ]';
		} else {
			push @autolink, $text
				if ($text ne $$img{title}); ## this image has same title
		}
		$href = mk_href($ii);
		$menu .= mk_link($href,$text) . "\n";
	}
	for my $aa (keys %alias) {
		next if $alias{$aa} eq $tt; ## ignore if refers to this image
		push @autolink, $aa;
	}
	$ptn = join('|', map { s/[^\w\s]/\\$&/g; $_ } @autolink);
	$ptn = qr/$ptn/;  ## ptn is like "Fred|Sam|Jane";

	# end of the menu
	#
	for my $aa (@menu_end) {
		$menu .= mk_link_plus($$aa{url},$$aa{title}) . "\n";
	}
	$menu .= '</div>';
	#printf STDERR "menu:%s\n\n", $menu if $debug;
	printf STDERR "tt:%s\n", $tt if $debug;

	# create intermediate image file
	my $mm = substr($big,rindex($big,'/')+1); # cheap version of "basename"
	my $big_ident = `identify $big`;
	$big_ident =~ /\s([0-9]+)x([0-9]+)[+\s]/
		or die "can't get size of '$big'";
	my($big_w,$big_h) = ($1,$2);
	my $mfile;
	if (($convert_geom_w >= $big_w) && ($convert_geom_h >= $big_h)) {
		if ($bigfiles) {
			$mfile = $big;
		} else {
			my @args = ( "/bin/cp",  $big, $mdir . $mm );
			printf STDERR "%s\n"
				, "copying $big to ${mdir}$mm" if $debug;
			system @args;
			die "Command '" . join(' ',@args) . "' failed: $!\n" if ($? >> 8);
			$mfile = "${mdir}$mm";
		}
	}
	elsif (!$no_create) {
		mkdir $mdir, 0755 if (! -d $mdir);
		my $rc = system("test ! -e ${mdir}$mm -o $big -nt ${mdir}$mm");
		die 'cannot run "test"' if $rc == -1;
		if ($rc >> 8 == 0) {
			my @args = (qw/convert -geometry/ ## ImageMagick
				, $convert_geom, $big, $mdir . $mm
				);
			printf STDERR "%s\n"
				, "creating ${mdir}$mm ($convert_geom)" if $debug;
			system @args;
			die "Command '" . join(' ',@args) . "' failed: $!\n" if ($? >> 8);
		}
		$mfile = "${mdir}$mm";
	}
	my $ident = `identify $mfile`; ## ImageMagick
	$ident =~ /\s([0-9]+)x([0-9]+)[+\s]/ or die "can't get size of '$mfile'";
	my($width,$height) = ($1,$2);
	printf STDERR "%sx%s\n", $width, $height if $debug;
	
	my $outfilename = $pgdir . "/" . $$img{pgfile};
	my $bigfile_url = 
		$bigfiles ?  $serve_dir . '/' . $big 
		: undef ;

	my $descr = $$img{html_description};
	$descr = subst_word($descr, \&title_to_link, $ptn);
	my @rrr = @{$$img{rings}};
	if (@rrr) {
		$descr .= $br . $br;
		for my $rr (@rrr) {
			my($ring,$index) = ($$rr{ring},$$rr{index});
			my $link_text = $$ring{text};
			my $r_images = $$ring{images}; # array ref
			die "ring messed up" if $$r_images[$index] != $img;
			my $r_next = ${$r_images}[($index+1) % @$r_images];
			my $href = mk_href($r_next);
			$descr .= mk_link($href, $link_text) . $br;
		}
	}

	my $subst_values = {
		pic_url =>  $serve_dir . '/' . $mfile
        ,pic_width => $width
        ,pic_height => $height
        ,pic_bigfile_url => $bigfile_url
		,pic_menu => $menu,
		,pic_menu_width => $menu_width,
        ,pic_name => $$img{title},
        ,pic_description => $descr,
        ,pic_copyright => $copyright
		};
	my $html = '';
	$html .= create_photo_page(
			$subst_values,
			$page_template,
			$img_template);
	if (!$page_template) {
		$html .= "\n";
		if ($complete_html) {
			$html = '<html><body>' . "\n"
				. $html . "\n"
				. '</body></html>' . "\n";
		}
	}
	
	my $fh;
	my $old_contents = '';
	eval { $old_contents = getfile($outfilename); };
	if ($old_contents ne $html) {
		if ($debug) {
			printf STDERR "%s: ", $outfilename;
			compare_lines($old_contents, $html);
		}
		printf STDERR "updating:%s\n", $outfilename if $debug;
		if (!open($fh, ">$outfilename" )) {
			my $msg = sprintf "%s\n", "can't open '$outfilename': $!";
			die $msg;
		}
		print $fh $html;
		close $fh;
	} else {
		printf STDERR "unchanged:%s\n", $outfilename if $debug;
	}
}

sub compare_lines {
	my($a,$b) = @_;
	my @la = split("\n",$a);
	my @lb = split("\n",$b);
	if (@la != @lb) {
		printf STDERR "linecounts: %d,%d\n", 0+@la,0+@lb if $debug;
		return;
	}
	for (my $i=0; $i<@la; $i++) {
		if ($la[$i] ne $lb[$i]) {
			printf STDERR "line %d differs\n", $i+1 if $debug;
			return;
		}
	}
} # end compare_lines

# Create the pages
my $pagelist_fh;
if (!open($pagelist_fh, ">$pagelist" )) {
	my $msg = sprintf "%s\n", "can't open '$pagelist': $!";
	die $msg;
}

# Add each image to its "rings" of images, if any
# An image is added to a ring if it matches
# the ring's pattern.
#
for my $img (@images) {
	for my $rr (@rings) {
		my $text = $$img{title};
		my $ptn = $$rr{pattern};
		if ($ptn =~ /^([\w]+)=(.*)/m ) {
			warn "no image field '$1'" if ! $$img{$1};
			if ($$img{$1}) {
				$text = $$img{$1};
				$ptn = $2;
			}
		}
		$ptn = qr/$ptn/;
		if ($text =~ $ptn) {
			my $iref = $$rr{images};
			push @$iref, $img;
			my $index = -1+@$iref;
			print STDERR "index=$index, img=$img\n";
			printf STDERR "images: %s\n", join(", ", @$iref);
			push @{$$img{rings}}, {
				ring=>$rr,
				index=>$index
				};
		}
	}
}

for my $img (@images) {
	createfiles($img);
	my $rec = mk_page_entry_text(
		 $$img{number}
		,$pgdir . '/' . $$img{pgfile}
		,$$img{title}
		,$$img{id}
		);
	printf $pagelist_fh "%s\n", $rec;
}
close $pagelist_fh;

